#include <cmath>
#include <cstdlib>
#include <memory>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <printf.hpp>
#include <hpc/cuda/check.h>
#include <hpc/cuda/copy.h>
#include <hpc/cuda/gematrix.h>
#include <hpc/matvec/gematrix.h>
#include "hsvcolor.hpp"
#include "rgbcolor.hpp"

using T = double;
using Index = std::size_t;
constexpr Index BLOCK_SIZE = 16;
constexpr Index MAX_GRID_SIZE = 32;

template<typename MA>
__global__ void init_matrix(MA A) {
   using T = typename MA::ElementType;

   Index i = threadIdx.y + blockIdx.y * blockDim.y;
   Index j = threadIdx.x + blockIdx.x * blockDim.x;

   const auto PI = std::acos(-T(1.0));
   const auto E = std::exp(T(1.0));
   const auto E_POWER_MINUS_PI = std::pow(E, -PI);

   Index N = A.numRows;
   T value;
   if (i == N-1) {
      value = std::sin(M_PI * (static_cast<T>(j)/N));
   } else if (i == 0) {
      value = std::sin(M_PI * (static_cast<T>(j)/N)) * E_POWER_MINUS_PI;
   } else {
      value = 0;
   }
   A(i, j) = value;
}

template<typename MA>
__global__ void init_matrix_border(MA A) {
   using T = typename MA::ElementType;

   Index index = threadIdx.x + blockIdx.x * blockDim.x;

   const auto PI = std::acos(-T(1.0));
   const auto E = std::exp(T(1.0));
   const auto E_POWER_MINUS_PI = std::pow(E, -PI);

   Index N = A.numRows;
   A(0, index) = std::sin(M_PI * (static_cast<T>(index)/N)) * E_POWER_MINUS_PI;
   A(N-1, index) = std::sin(M_PI * (static_cast<T>(index)/N));
   if (index > 0 && index < N-1) {
      A(index, 0) = 0; A(index, N-1) = 0;
   }
}

template<typename MA, typename MB>
__global__ void jacobi_iteration(const MA A, MB B) {
   using T = typename MA::ElementType;

   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;

   T value;
   if (i == 0 || j == 0 || i == A.numRows-1 || j == A.numCols-1) {
      value = A(i, j);
   } else {
      value = 0.25 *
	 (A(i-1, j) + A(i, j-1) + A(i, j+1) + A(i+1, j));
   }
   B(i, j) = value;
}

template<typename MA, typename MB, typename MDiff>
__global__ void compute_local_max_diff(const MA A, const MB B,
      MDiff global_diff) {
   using T = typename MA::ElementType;
   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;

   __shared__ T diff[BLOCK_SIZE][BLOCK_SIZE];
   diff[threadIdx.y][threadIdx.x] = std::abs(A(i, j) - B(i, j));

   /* find maximum of diff[i][j] within a block */
   unsigned int offset = BLOCK_SIZE / 2;
   while (offset > 0) {
      __syncthreads();
      if (threadIdx.x < offset) {
	 T other_diff = diff[threadIdx.y][offset + threadIdx.x];
	 if (other_diff > diff[threadIdx.y][threadIdx.x]) {
	    diff[threadIdx.y][threadIdx.x] = other_diff;
	 }
      }
      if (threadIdx.y < offset) {
	 T other_diff = diff[offset + threadIdx.y][threadIdx.x];
	 if (other_diff > diff[threadIdx.y][threadIdx.x]) {
	    diff[threadIdx.y][threadIdx.x] = other_diff;
	 }
      }
      offset /= 2;
   }
   if (threadIdx.x == 0 && threadIdx.y == 0) {
      global_diff(blockIdx.y, blockIdx.x) = diff[0][0];
   }
}

__device__ T global_max_diff; // result of compute_global_max_diff

template<typename MDiff>
__global__ void compute_global_max_diff(MDiff global_diff) {
   using T = typename MDiff::ElementType;
   unsigned int i = threadIdx.y;
   unsigned int j = threadIdx.x;
   __shared__ T diff[MAX_GRID_SIZE][MAX_GRID_SIZE];
   diff[i][j] = global_diff(i, j);
   unsigned int offset = gridDim.x / 2;
   while (offset > 0) {
      __syncthreads();
      if (j < offset) {
	 T other_diff = diff[i][j + offset];
	 if (other_diff > diff[i][j]) {
	    diff[i][j] = other_diff;
	 }
      }
      if (i < offset) {
	 T other_diff = diff[i + offset][offset];
	 if (other_diff > diff[i][j]) {
	    diff[i][j] = other_diff;
	 }
      }
      offset /= 2;
   }
   if (i == 0 && j == 0) {
      global_max_diff = diff[0][0];
   }
}

template<typename MA, typename MB>
__global__ void blowup(const MA A, MB B, unsigned int log2factor) {
   using T = typename MA::ElementType;
   /* indices of target matrix B */
   Index B_i = threadIdx.y + blockIdx.y * blockDim.y;
   Index B_j = threadIdx.x + blockIdx.x * blockDim.x;
   /* indices of target matrix A */
   Index A_i = B_i >> log2factor;
   Index A_j = B_j >> log2factor;
   B(B_i, B_j) = A(A_i, A_j);
}

template<typename MA>
__global__ void color_converter(const MA A, guchar* pixels,
      unsigned int rowstride, unsigned int n_channels) {
   using T = typename MA::ElementType;
   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;
   T value = A(i, j);
   float hue = (1 - value) * 240;
   HSVColor hsv_color(hue, 1, 1);
   RGBColor rgb_color = hsv_color.get_rgb();

   guchar* pp = pixels + i * rowstride + j * n_channels;
   pp[0] = rgb_color.red * 255;
   pp[1] = rgb_color.green * 255;
   pp[2] = rgb_color.blue * 255;
}

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   if (argc != 1) {
      std::cerr << "Usage: " << cmdname << " out.png" << std::endl;
      exit(1);
   }

   const char* outfile = *argv++; --argc;

   constexpr Index MAX_N = MAX_GRID_SIZE * BLOCK_SIZE;
   GdkPixbuf* pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
      MAX_N, MAX_N);
   int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
   int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
   guchar* pixels = gdk_pixbuf_get_pixels(pixbuf);
   if (!pixels) {
      std::cerr << cmdname << ": failed to allocate pixbuf with " <<
	 MAX_N << " x " << MAX_N << " pixels" << std::endl;
      exit(1);
   }

   unsigned int iterations = 0;
   hpc::cuda::GeMatrix<T, Index> A_dev(MAX_N, MAX_N, hpc::cuda::RowMajor);
   hpc::cuda::GeMatrix<T, Index> B_dev(MAX_N, MAX_N, hpc::cuda::RowMajor);
   for (Index grid_size = 1; grid_size <= MAX_GRID_SIZE; grid_size <<= 1) {
      hpc::cuda::GeMatrix<T, Index> global_diff_dev(grid_size, grid_size,
	 hpc::cuda::RowMajor);
      auto global_diff = global_diff_dev(0, 0, grid_size, grid_size);
      dim3 block_dim(BLOCK_SIZE, BLOCK_SIZE);
      dim3 grid_dim(grid_size, grid_size);
      Index N = grid_size * BLOCK_SIZE;
      auto A = A_dev(0, 0, N, N);
      auto B = B_dev(0, 0, N, N);
      if (iterations == 0) {
	 init_matrix<<<grid_dim, block_dim>>>(A);
      } else {
	 auto last = A_dev(0, 0, N/2, N/2);
	 blowup<<<grid_dim, block_dim>>>(last, B, 1);
	 init_matrix_border<<<1, N>>>(B);
	 jacobi_iteration<<<grid_dim, block_dim>>>(B, A); ++iterations;
      }
      for(;;) {
	 for(unsigned int i = 0; i < 10; ++i) {
	    jacobi_iteration<<<grid_dim, block_dim>>>(A, B); ++iterations;
	    jacobi_iteration<<<grid_dim, block_dim>>>(B, A); ++iterations;
	 }
	 compute_local_max_diff<<<grid_dim, block_dim>>>(A, B, global_diff);
	 compute_global_max_diff<<<1, grid_dim>>>(global_diff);
	 T maxdiff;
	 cudaMemcpyFromSymbol(&maxdiff, global_max_diff, sizeof(T));
	 fmt::printf("grid_size = %2d, iterations = %6d, maxdiff = %8g\r",
	    grid_size, iterations, maxdiff);
	 if (maxdiff < 1E-7) break;
      }
      fmt::printf("\n");
   }

   guchar* pixels_dev;
   size_t pixels_size = MAX_N * rowstride * sizeof(guchar);
   CHECK_CUDA(cudaMalloc, (void**)&pixels_dev, pixels_size);
   dim3 block_dim(BLOCK_SIZE, BLOCK_SIZE);
   dim3 grid_dim(MAX_GRID_SIZE, MAX_GRID_SIZE);
   auto A = A_dev(0, 0, MAX_N, MAX_N);
   color_converter<<<grid_dim, block_dim>>>(A,
      pixels_dev, rowstride, n_channels);
   CHECK_CUDA(cudaMemcpy, pixels, pixels_dev, pixels_size,
      cudaMemcpyDeviceToHost);

   gdk_pixbuf_save(pixbuf, outfile, "png", (GError **)0, (void*)0);
}
